--[[
# Copyright 2020 Quarq
# This file is part of A-Bomb: Azeroth Bombardier.
# A-Bomb: Azeroth Bombardier is distributed under a BSD License.
# It is provided AS-IS and all warranties, express or implied, including, but not
# limited to, the implied warranties of merchantability or fitness for a particular
# purpose, are disclaimed.  See the LICENSE file for full information.
--]]
local REALNAME, _ = ...
local MYNAME = GetAddOnMetadata(REALNAME, "X-ShortName") or REALNAME
local Addon = CreateFrame("Frame", MYNAME, UIParent)

Addon:SetScript("OnEvent", function(self,event,...) if (Addon[event]) then Addon[event](self, ...) end end)

Addon:RegisterEvent("PLAYER_LOGIN")


local HBD = LibStub("HereBeDragonsQuestie-2.0")


local GW, GH = 1/20, 1.5/20
Addon.GW = GW
Addon.GH = GH

local MW, MH = WorldMapFrame:GetCanvas():GetWidth(), WorldMapFrame:GetCanvas():GetHeight()

local EXPLOSION_RADIUS = 14 / MW

local DENSITY = 5	-- we'll draw at most DENSITY x DENSITY targets in a grid


-- make local references to globals (faster)
local floor, abs, atan2 = math.floor, math.abs, math.atan2
local function normangle(angle) angle=angle+4*PI; angle=angle-2*PI*floor(angle/(2*PI)); return angle<=PI and angle or angle-2*PI; end
local function min(a,b) return (a<b and a or b) end
local function max(a,b) return (a>b and a or b) end
local tinsert, tremove = tinsert, tremove
local sprintf = string.format

local mapdistance2 -- forward -- see PLAYER_LOGIN


function Addon : PLAYER_LOGIN(arg)

	self.realm = GetRealmName()
	self.player = UnitName("player")
	self.class = string.upper(UnitClass("player"))

	-- initialize settings database
	ABombData = ABombData or {}
	-- ABombData[self.player] = ABombData[self.player] or {}
	-- Addon.DB = ABombData[self.player]
	Addon.DB = ABombData

	if (Addon.DB.automap == nil) then
	end

	self:InitRanks()

	-- define slash commands... the array index must match the substring in the global variables
	SlashCmdList["ABOMB"] = function(msg) self:Slash(msg) end;
	--    v------^
	SLASH_ABOMB1= "/bomb";

	self:UnregisterEvent("PLAYER_LOGIN")

	self:RegisterEvent("PLAYER_CONTROL_LOST")
	self:RegisterEvent("PLAYER_CONTROL_GAINED")

	self:RegisterEvent("GOSSIP_SHOW")
	self:RegisterEvent("GOSSIP_CLOSED")

	if (Addon.DB.enabled == nil) then
		self:Slash("on")
		self:Slash("automap on")
		print(Addon.gold("Visit a flightmaster for your first mission."))
	end

	mapdistance2 = Addon.mapdistance2
end


function Addon : Slash(msg)
	msg = msg:trim()

	if (msg == "") then
		local version = GetAddOnMetadata(REALNAME, "Version")
		print(Addon.gold("Azeroth Bombardier (v"..version..") commands:"))
		print("on/off", Addon.gold("turns the game on or off"))
		print("automap on/off", Addon.gold("whether to open the map automatically when accepting a mission"))
		print("rank", Addon.gold("show your rank, stats, and promotion progress"))
		return
	end


	if (msg == "rank") then
		local version = GetAddOnMetadata(REALNAME, "Version")
		print(Addon.gold("Azeroth Bombardier (ver. "..version..")"))
		local rank = Addon:GetCurrentRank()
		print("Your rank is", Addon.gold(rank.name))
		print("  ", "Neutralized", Addon.commaformat(Addon.DB.total_kills), "targets")
		print("  ", "Completed", Addon.commaformat(Addon.DB.total_missions), "missions")
		print("  ", "Total sorties:", Addon.commaformat(Addon.DB.total_sorties))
		if (rank.promotion) then
			local nextrank = Addon:GetNextRank()
			print("For promotion to", nextrank.name .. ":")
			print("  ", "Neutralize", Addon.commaformat(rank.promotion.kills), "targets")
			print("  ", "Complete", Addon.commaformat(rank.promotion.missions), "missions")
			print("  ", "Complete", nextrank.name, "challenge mission")
		end
		return
	end

	if (msg == "on") then
		Addon.DB.enabled = true
		local rank = Addon:GetCurrentRank()
		print(Addon.gold("Azeroth Bombardier is on.  Your rank is " .. rank.name .. "."))
		return
	elseif (msg == "off") then
		if (Addon.DB.enabled) then
			Addon.DB.enabled = false
			self:SetScript("OnUpdate", nil)
			self:EndGame()
			self:SetZone(nil)
			if (self.mission) then
				print(Addon.gold("Azeroth Bombardier:  Mission aborted."))
			end
		end
		local rank = Addon:GetCurrentRank()
		print(Addon.gold("Azeroth Bombardier is off.  Your rank is " .. rank.name .. "."))
		return
	end

	if (msg:match("^automap ")) then
		local flag = msg:match("^automap (%a+)") or ""
		if (flag:lower() == "on") then
			Addon.DB.automap = true
			print(Addon.gold("Your map will open AUTOMATICALLY when you accept a bombing mission."))
		elseif (flag:lower() == "off") then
			Addon.DB.automap = false
			print(Addon.gold("You must open your map MANUALLY after you accept a bombing mission."))
		end
		return
	end

end


function Addon : GOSSIP_SHOW()

	-- if report left over from last time, destroy it
	if (self.report) then
		self.report:SetScript("OnHide", nil)
		self.report:Hide()
		self.report:ClearAllPoints()
		self.report:Clear() -- contents
		self.report = nil
	end

	if (not self.DB.enabled) then return end

	-- if it's flight-master gossip, then I want to offer a mission
	self.taxi_option = nil
	local options = { GetGossipOptions() }
	local n = GetNumGossipOptions()
	for o = 1, n do
		-- local otext = options[o*2-1]
		local otype = options[o*2]
		if (otype == "taxi") then
			self.taxi_option = o
			break
		end
	end

	-- if there's a taxi option, then show the mission briefing
	if (self.taxi_option) then
		self.briefing = self:NewMissionBriefing()
		self.briefing:SetParent(GossipGreetingScrollChildFrame)
		self.briefing:SetPoint("BOTTOMLEFT", GossipGreetingScrollChildFrame, "BOTTOMLEFT", 9, 120)
		self.briefing:Show()

		self.mission_accepted = nil	-- set to TRUE if "accept mission" button pressed
	end
end


function Addon : GOSSIP_CLOSED()
	-- done with the briefing
	if (self.briefing) then
		self.mission = self.briefing.mission	-- nil at max rank
		self.briefing:Hide()
		self.briefing:ClearAllPoints()
		self.briefing:Clear() -- content
		self.briefing = nil
	end
end


function Addon : PLAYER_CONTROL_LOST()
	if (not self.DB.enabled) then return end
	if (not self.mission_accepted) then return end
	print("mission was accepted")

	self:SetScript("OnUpdate", self.OnUpdate)
	self.elapsed = 0
end


function Addon : PLAYER_CONTROL_GAINED()
	self:SetScript("OnUpdate", nil)
	self.mission_accepted = nil

	if (self.playing) then
		self:EndGame()
		assert(not self.playing)
		assert(not self.flying)
	end
end


function Addon : StartGame()
	if (not self.DB.enabled) then return end


	self:SetZone( WorldMapFrame:GetMapID() or C_Map.GetBestMapForUnit("player") )

	if (self.sight == nil) then
		self.sight = self:MakeBombSight()
		assert(self.sight)
	end

	self.targets = {}
	self.killed = {}

	-- counts
	self.kills = 0
	self.dropped = 0

	if (self.mission) then
		print(Addon.gold("Azeroth Bombardier: Mission Starting!"))
	end
	self.status = self:NewMissionStatus(self.mission)	-- mission nil at max rank
	self.status:SetParent(WorldMapFrame.ScrollContainer)
	self.status:SetPoint("TOPRIGHT", -4, -1)
	self.status:Show()

	self:ShowAll()
end


function Addon : EndGame()
	if (not self.playing) then return end

	self.DB.total_sorties = self.DB.total_sorties + 1

-- local killed = self.mission.color and self.status.kills[self.mission.color] or self.status.totalkills
-- local survivors = self.status.survivors / (killed + self.status.survivors)
-- print("Killed", killed, (self.mission.color and (self.mission.color.." left") or "left"), self.status.survivors, sprintf("(%.2f%%)", 100*survivors))

	self.report = self:NewMissionReport(self.mission, self.status)	-- mission nil at max rank
	self.report:SetParent(WorldMapFrame.ScrollContainer)
	self.report:SetPoint("CENTER")
	self.report:Show()

	-- don't clear the zone (removing corpses, among other important things) until the map is closed (hiding the report)
	self.report:SetScript("OnHide", function() Addon:SetZone(nil) end)

	self.status:Hide()
	self.status:ClearAllPoints()
	self.status = nil

	self.mission = nil

	self:DestroyBombSight(self.sight)
	self.sight = nil

	if (self.bomb) then
		self.bomb:Abort() -- cancel any in-progress sounds
		self:RemoveBomb(self.bomb)
		self.bomb = nil
	end

	-- destroy (living) targets
	for key, target in pairs(self.targets) do
		self:RemoveTarget(target)
	end
	self.targets = {}

	self.flying = false
	self.playing = false
end


function Addon : ShowAll(show)
	if (show == nil) then show=true end

	if (self.bomb) then
		self.bomb:Abort()
		self:RemoveBomb(self.bomb)
		self.bomb = nil
	end

	if (self.sight) then
		if (show) then self.sight:Show() else self.sight:Hide() end
	end

	for _, target in pairs(self.targets) do
		if (show) then target:Show() else target:Hide() end
	end

	for _, corpse in pairs(self.killed) do
		if (show) then corpse:Show() else corpse:Hide() end
	end
end


function Addon : HideAll()
	self:ShowAll(false)
end


local WORLD, KALIMDOR, EASTERN = 947, 1414, 1415

function Addon : SetZone(zone)

	if (self.report) then
		self.report:SetScript("OnHide", nil)
		self.report:Hide()
		self.report:ClearAllPoints()
		self.report:Clear() -- contents
		self.report = nil
	end

	if (zone == WORLD or zone == KALIMDOR or zone == EASTERN) then
		zone = nil
		if (self.sight) then self.sight:Hide() end
	end

	if (zone == self.zone) then return end
	-- print("new map zone", zone)

	if (zone == nil) then
		self.zone = zone
		self.gridkey = nil
		self.mobs = {}
		self:RemoveAllTargets()
		self.targets = {}
		self.killed = {}
	else
		local oldzone = self.zone

		self.zone = zone
		self.gridkey = nil

		self.mobs = self.mobs or {}
		self.targets = self.targets or {}
		self.killed = self.killed or {}

		-- correct mx,my of mobs in the new zone, and discard other mobs
		for mz, mobs in pairs(self.mobs) do
			if (mz == self.zone) then
				for _, mob in pairs(mobs) do
					mob.mx, mob.my = mob.x, mob.y
					mob.mz = self.zone
				end
			else
				self.mobs[mz] = nil
			end
		end

		if (oldzone) then
			-- correct position of targets in the new zone, and discard other targets
			local outgoing = self.targets
			self.targets = {}
			for _, target in pairs(outgoing) do
				if (target.zone == self.zone) then
					local x,y = target:GetPosition()
					x,y = HBD:TranslateZoneCoordinates(x, y, oldzone, self.zone, true)
					target:SetPosition(x,y)
					local key = sprintf("Z%d:%03d:%03d", target.zone, floor(DENSITY*x/GW), floor(DENSITY*y/GH))
					self.targets[key] = target
				else
					self:RemoveTarget(target)
				end
			end

			-- correct position of corpses in the new zone, and discard other targets
			outgoing = self.killed
			self.killed = {}
			for key, corpse in pairs(outgoing) do
				if (corpse.zone == self.zone) then
					local x,y = corpse:GetPosition()
					x,y = HBD:TranslateZoneCoordinates(x, y, oldzone, self.zone, true)
					corpse:SetPosition(x,y)
					local key = sprintf("Z%d:%03d:%03d", corpse.zone, floor(DENSITY*x/GW), floor(DENSITY*y/GH))
					self.killed[key] = corpse
				else
					self:RemoveTarget(corpse)
				end
			end
		end

		if (C_Map.GetPlayerMapPosition(self.zone, "player")) then
			self.auto_zone_change_ok = true
		else
			self.auto_zone_change_ok = false
		end
	end
end


function Addon : OnUpdate(dt)
	if (not self.flying) then
		self.flying = UnitOnTaxi("player")
		if (self.flying and self.mission_accepted and Addon.DB.automap and not WorldMapFrame:IsShown()) then
			print("auto-open world map")
			ShowUIPanel(WorldMapFrame)
		end
	end

	if (not self.playing) then
		self.playing = self.flying and WorldMapFrame:IsShown()
		if (self.playing) then
			self:StartGame()
			-- and fall through
		else
			return
		end
	end
	assert(self.playing)

	if not (self.zone and WorldMapFrame:IsShown()) then
		return
	end

	-- see whether the player has crossed into a zone that I should be interested in
	-- that is, a zone I can get a position for, and that I haven't seen yet
	local pz = C_Map.GetBestMapForUnit("player")
	if (pz and not self.mobs[pz]) then
		-- haven't seen it yet...  can I get a MZ position for player in PZ ?
		local pos = C_Map.GetPlayerMapPosition(self.zone, "player")
		if (pos) then
			-- print("new interesting zone", pz)
			-- this PZ is interesting to me
			local x,y  = pos:GetXY()
			x, y = HBD:TranslateZoneCoordinates(x, y, self.zone, pz, true)
			assert(x and y)
			self.mobs[pz] = self:FetchMobs(pz, x, y)
			assert(self.mobs[pz])
		end
	end

	-- need the player position on the current map
	local pos = C_Map.GetPlayerMapPosition(self.zone, "player")

	-- if not available, then it may be because he's moved onto a different map
	if (pos == nil and self.auto_zone_change_ok) then
		local best = C_Map.GetBestMapForUnit("player")
		if (best) then
			WorldMapFrame:SetMapID(best)
			assert(self.zone == best)	-- via OnMapChanged() notification

			pos = C_Map.GetPlayerMapPosition(best, "player")
			if (pos == nil) then
				-- print("no position in new best", best)
				return
			end
			-- else pos is ok and fall through
		else
			-- print("no best zone")
			return
		end
	end

	if (pos) then
		local px, py = pos:GetXY()

		-- position and orient the bomb sight
		if (not self.sight:IsShown()) then self.sight:Show() end
		local facing = GetPlayerFacing()	-- counter-clockwise from North
		self.sight:SetRotation(facing)
		self.sight:SetPosition(px,py)

		-- if he's crossed into a new grid, get the fresh set of nearby mobs
		local gridkey = sprintf("%02d:%02d", floor(px/GW), floor(py/GH))
		if not (gridkey == self.gridkey) then
			self.gridkey = gridkey
			-- print("new grid", gridkey)

			-- get new lists of mobs for each zone we care about
			for zone, _ in pairs(self.mobs) do
				local x, y = HBD:TranslateZoneCoordinates(px, py, self.zone, zone, true)
				self.mobs[zone] = self:FetchMobs(zone, x, y)
				assert(self.mobs[zone])
			end
		end

		-- less frequently...
		self.elapsed = self.elapsed + dt
		if (self.elapsed >= 0.10) then
			self.elapsed = self.elapsed - 0.10

			-- update actual targets
			local radius = self.sight:GetWidth()/2 / MW
			local radius2 = radius * radius

			-- remove targets that have passed out of range
			local include = self.mission and self.mission.color or nil		-- nil mission OK (nil inclusion color)
			local exclude = self.mission and self.mission.exclude or nil	-- nil mission OK (nil exclusion color)
			for key, target in pairs(self.targets) do
				local tx, ty = target:GetPosition()
				if (mapdistance2(tx,ty, px,py) > radius2) then
					self:RemoveTarget(target)
					self.targets[key] = nil

					-- if the mission has a color criteria, only count survivors of that color
					if (target.color.name == exclude) then
						-- do NOT count it as a survivor
					elseif (target.color.name == include or include == nil) then
						-- either no mission color criteria, or this target matches it, so we count it as a survivor
						self.status.survivors = self.status.survivors + 1
					end
				end
			end

			-- add targets that have moved into range
			local nColors = #Addon.COLORS
			for zone, mobs in pairs(self.mobs) do
				for _, mob in pairs(mobs) do
					if (mapdistance2(mob.mx, mob.my, px, py) <= radius2) then
						local key = sprintf("Z%d:%03d:%03d", zone, floor(DENSITY*mob.mx/GW), floor(DENSITY*mob.my/GH))
						if not (Addon.targets[key] or Addon.killed[key]) then
							-- self.targets[key] = self:DrawTarget(mob.mx, mob.my, zone, mob.npcid/7)
							self.targets[key] = self:DrawTarget(mob.mx, mob.my, zone, Addon.COLORS[1+mob.npcid%nColors])
						end
					end
				end
			end
		end -- less frequently

		if (self.bomb) then
			if (self.bomb.detonated) then
				self.bomb.detonated = false

				-- now it counts as "used"
				self.status.bombs = self.status.bombs + 1

				local bx,by = self.bomb:GetPosition()

				local killed = {}
				for key, target in pairs(self.targets) do
					local tx,ty = target:GetPosition()
					if (mapdistance2(bx,by, tx,ty) <= EXPLOSION_RADIUS*EXPLOSION_RADIUS) then
						tinsert(killed, key)
					end
				end
				if (#killed >= 4) then
					self.bomb:ShowCombo(#killed)
				end
				while (killed[1]) do
					local key = tremove(killed)
					local target = self.targets[key]

					target:Die()
					assert(target.dead)
					self.kills = self.kills + 1

					self.status.totalkills = self.status.totalkills + 1
					self.status.kills[target.color.name] = (self.status.kills[target.color.name] or 0) + 1

					self.targets[key] = nil
					self.killed[key] = target
				end

				self.status:UpdateProgress()

				-- doing this here allows the next one to be dropped before this one's explosion finishes
				self.bomb = nil
			end

			-- if (self.bomb.finished) then
			-- 	self:RemoveBomb(self.bomb)
			-- 	self.bomb = nil
			-- end
		end

	else

		if (self.sight:IsShown()) then self.sight:Hide() end

	end -- pos non-nil
end


function Addon : FetchMobs(zone, x, y)
	-- print("fetching mobs near", zone, x, y)
	local mobs = Addon:GetMobsNear(zone, x, y)
	-- they come back with properties specific to their own zone...
	-- give each one a position on MY map
	for _, mob in pairs(mobs) do
		x, y = HBD:TranslateZoneCoordinates(mob.x, mob.y, zone, self.zone, true)
		mob.mz = self.zone
		mob.mx = x
		mob.my = y
	end
	return mobs
end


--
--  Bomb Sight marker / data-provider / pin-pool implementation
--

-- my marker
local BombSight = CreateFromMixins(MapCanvasPinMixin)

local _provider = CreateFromMixins(MapCanvasDataProviderMixin)
WorldMapFrame:AddDataProvider(_provider)

function _provider : OnMapChanged()
	Addon:SetZone( self:GetMap():GetMapID() )
end

local POOLNAME = "ABombSight"

local _pool = CreateFramePool("FRAME")
WorldMapFrame.pinPools[POOLNAME] = _pool

function _pool.creationFunc(pool)
    return Mixin(CreateFrame("FRAME"), BombSight)
end

function _pool.resetterFunc(pool, pin)
    FramePool_HideAndClearAnchors(pool, pin)
    pin:OnReleased()
	-- these were assigned by MapCanvas:AcquirePin
    pin.pinTemplate = nil
    pin.owningMap = nil
end


function BombSight : OnAcquired(x, y)

	self:SetPosition(x, y)
	self:SetSize(128, 128)

	if (not self.texture) then
		self.texture = self:CreateTexture(nil, "OVERLAY", nil, 0)
		self.texture:SetTexture("Interface/Addons/A-Bomb/textures/sight")
		self.texture:SetAllPoints(self)

		self:SetScript("OnEnter", self.OnMouseEnter)
		self:SetScript("OnLeave", self.OnMouseLeave)
		self:SetScript("OnMouseDown", self.OnClick)
	end

    self:UseFrameLevelType("PIN_FRAME_LEVEL_AREA_POI")
end


function BombSight : OnMouseEnter()
end


function BombSight : OnMouseLeave()
end


function BombSight : OnClick(which)
	if (which == "LeftButton" and Addon.zone and not Addon.bomb) then
	    local mx, my = Addon.get_mouse_position_map()

		local position = C_Map.GetPlayerMapPosition(Addon.zone, "player")
		local px, py = position:GetXY()

		local radius = self:GetWidth()/2 / MW

		if (mapdistance2(mx,my, px,py) <= radius*radius) then
			Addon.bomb = Addon:DrawBomb(px,py, mx,my)
		end
	end
end


function BombSight : OnReleased()
	-- nothing
end


function BombSight : SetRotation(radians)
	self.texture:SetRotation(radians)
end


function Addon : MakeBombSight()
 	return _provider:GetMap():AcquirePin(POOLNAME, 0.50, 0.50)
end

function Addon : DestroyBombSight(sight)
	_provider:GetMap():RemovePin(sight)
end
