--[[
# Copyright 2020 Quarq
# This file is part of A-Bomb: Azeroth Bombardier.
# A-Bomb: Azeroth Bombardier is distributed under a BSD License.
# It is provided AS-IS and all warranties, express or implied, including, but not
# limited to, the implied warranties of merchantability or fitness for a particular
# purpose, are disclaimed.  See the LICENSE file for full information.
--]]
local MYNAME, _ = ...
MYNAME = GetAddOnMetadata(MYNAME, "X-ShortName") or MYNAME
local Addon = _G[MYNAME]
assert(Addon)

local sprintf = string.format
local floor = math.floor


local FONTSIZE = 13
local MARGIN = FONTSIZE	-- side margins
local V_MARGIN = math.floor(FONTSIZE/2)


local MissionFrame = {}


function MissionFrame : Clear()
	local label
	for _, line in pairs(self.lines or {}) do
		label = line.label2; line.label2 = nil
		if (label) then
			label:Hide()
			label:SetText(nil)
			label:ClearAllPoints()
			label = nil
		end
		if (line.swatch) then
			line.swatch:Hide()
			line.swatch:SetTexture(nil)
			line.swatch:ClearAllPoints()
			line.swatch = nil
		end
		label = line.colorlabel; line.colorlabel = nil
		if (label) then
			label:Hide()
			label:SetText(nil)
			label:ClearAllPoints()
			label = nil
		end
		label = line.label1; line.label1 = nil
		if (label) then
			label:Hide()
			label:SetText(nil)
			label:ClearAllPoints()
			label = nil
		end
		if (line.icon) then
			line.icon:Hide()
			line.icon:SetTexture(nil)
			line.icon:ClearAllPoints()
			line.icon = nil
		end
	end

	if (self.accept) then
		self.accept:Hide()
		self.accept:SetScript("OnClick", nil)
		self.accept:SetText(nil)
		self.accept:ClearAllPoints()
		self.accept = nil
	end
	if (self.change) then
		self.change:Hide()
		self.change:SetScript("OnClick", nil)
		self.change:SetText(nil)
		self.change:ClearAllPoints()
		self.change = nil
	end

	self:SetSize(0,0)
	self.contentWidth = 0
	self.contentHeight = 0

	self.lines = {}
end



function MissionFrame : newlabel(size)
	local label = self:CreateFontString(nil, "ARTWORK")
	label:SetFont("Fonts/FRIZQT__.TTF", size)
	label:SetTextColor(0,0,0, 1)

	if (self.shadows) then
		label:SetShadowOffset(1,-1)
		label:SetShadowColor(1,1,1, 0.30)
	end

	return label
end


function MissionFrame : AddLine(text, size)
	local size = size or FONTSIZE

	local line = {}
	tinsert(self.lines, line)

	line.label1 = self:newlabel(size)
	line.label1:SetText(text)

	line.label1:SetPoint("TOPLEFT", self, "TOPLEFT", MARGIN, -V_MARGIN-self.contentHeight)
	self.contentHeight = self.contentHeight + size + 2

	self:SetHeight(V_MARGIN + self.contentHeight + V_MARGIN)

	return line
end


function MissionFrame : AddGap(height)
	self.contentHeight = self.contentHeight + FONTSIZE*height
end


function MissionFrame : AddObjective(icon, text)
	local before = self.contentHeight
	local line = self:AddLine(text)

	line.icon = self:CreateTexture(nil, "ARTWORK")
	line.icon:SetTexture("Interface/Addons/A-Bomb/textures/"..icon)
	line.icon:SetSize(FONTSIZE+2, FONTSIZE+2)

	line.icon:SetPoint("TOPLEFT", self, "TOPLEFT", MARGIN, -V_MARGIN-before+1)

	line.label1:ClearAllPoints()
	line.label1:SetPoint("LEFT", line.icon, "RIGHT", 3, 0)

	return line
end


function MissionFrame : AddObjectiveWithColor(icon, textBefore, colorname, textAfter)
	local line = self:AddObjective(icon, textBefore)

	-- local label

	local color = Addon.COLORS_BY_NAME[colorname]
	local r,g,b = Addon.hsv(color.hue, 1, 1)
	line.swatch = self:CreateTexture(nil, "ARTWORK", nil, -1)
	line.swatch:SetTexture("Interface/Addons/A-Bomb/textures/target-outlined")
	line.swatch:SetSize(FONTSIZE*9/13, FONTSIZE*9/13)
	line.swatch:SetVertexColor(r,g,b, 1)
	line.swatch:SetPoint("LEFT", line.label1, "RIGHT", 5, 0)

	-- label = self:CreateFontString(nil, "ARTWORK", nil, 0)
	-- label:SetFont("Fonts/FRIZQT__.TTF", FONTSIZE)
	-- label:SetTextColor(0,0,0, 1)
	-- label:SetText(colorname:lower())
	line.colorlabel = self:newlabel(FONTSIZE)
	line.colorlabel:SetText(colorname:lower())
	line.colorlabel:SetPoint("LEFT", line.swatch, "RIGHT", 2, 0)

	-- label = self:CreateFontString(nil, "ARTWORK", nil, 0)
	-- label:SetFont("Fonts/FRIZQT__.TTF", FONTSIZE)
	-- label:SetTextColor(0,0,0, 1)
	-- label:SetText(textAfter)
	line.label2 = self:newlabel(FONTSIZE)
	line.label2:SetText(textAfter)
	line.label2:SetPoint("LEFT", line.colorlabel, "RIGHT", 4, 0)

	return line
end


function MissionFrame : AddMissionButtons()
	local button

	button = CreateFrame("BUTTON", nil, self, "UIPanelButtonTemplate")
	button:SetSize(120, FONTSIZE+6)
	button:SetText("Accept Mission")
	button:SetPoint("TOPLEFT", self, "TOPLEFT", MARGIN, -V_MARGIN-self.contentHeight)
	self.accept = button
	self.accept:SetScript("OnClick", function() Addon.mission_accepted=true; SelectGossipOption(Addon.taxi_option) end)

	local rank = Addon:GetCurrentRank()
	if (rank.level == Addon.MAX_RANK) then
		local tags = { "Let's kick this pig", "Let's light this candle", "They're not gonna bomb themselves" }
		local roll = math.random(1,100)
		if (roll <= 3 * #tags) then
			self.accept:SetText(tags[1+roll%3])
		else
			self.accept:SetText("Let's go!")
		end
		self.accept:SetWidth(self.accept:GetFontString():GetUnboundedStringWidth()+30)
	else
		button = CreateFrame("BUTTON", nil, self, "UIPanelButtonTemplate")
		button:SetSize(120, FONTSIZE+6)
		button:SetText("Change Mission")
		-- button:SetPoint("TOPRIGHT", self, "TOPRIGHT", -MARGIN, -V_MARGIN-self.contentHeight)
		button:SetPoint("LEFT", self.accept, "RIGHT", MARGIN, 0)
		self.change = button
		local this = self
		self.change:SetScript("OnClick", function() this:OnChangeMission() end)
	end
end


function MissionFrame : GetContentWidth()
	local cw = 0
	for _, line in pairs(self.lines) do
		assert(line.label1)
		local w = line.label1:GetWidth()

		if (line.icon) then
			w = w + 3 + line.icon:GetWidth()
		end

		if (line.swatch) then
			w = w + 5 + line.swatch:GetWidth()
			assert(line.colorlabel)
			w = w + 2 + line.colorlabel:GetWidth()
			assert(line.label2)
			w = w + 4 + line.label2:GetWidth()
		end

		-- if (self.foo) then print(cw, "?", w, "=", line.label1:GetText()) end
		cw = math.max(cw, w)
	end
	return cw
end


--
--  Briefing
--

local MissionBriefing = Mixin({}, MissionFrame)


function Addon : NewMissionBriefing()
	local mf = Mixin(CreateFrame("FRAME"), MissionBriefing)
	mf.shadows = false
	local mission, more = Addon:GetMission()
	mf:SetMission(mission)
	if (mf.change and not more) then
		mf.change:SetEnabled(false)
	end
	return mf
end


function MissionBriefing : SetMission(mission)
	assert(not self.bg)

	MARGIN = FONTSIZE
	V_MARGIN = math.floor(FONTSIZE/2)

	self:Clear()

	local rank = Addon:GetCurrentRank()
	if (rank.level == Addon.MAX_RANK) then

		self:AddLine("Good hunting, " .. rank.name .. "!")
		self:AddGap(1.5)

		self:AddMissionButtons()

	elseif (mission) then

		self.mission = mission

		--[[
			Azeroth Bombardier

			Mission Briefing N-48/C                ("N" for Nuisance... T-17/E, B-35/L)
			[] Kill 25 [RED] targets
			[] Leave 25% or fewer alive
			[] Use no more than 15 bombs
			[] Avoid [GREEN] targets
		--]]

		self:AddLine("Mission Briefing:")
		if (mission.title) then
			self:AddLine(mission.title)
		end
		self:AddGap(0.50)

		if (mission.kills) then
			if (mission.color) then
				self:AddObjectiveWithColor("box", "Kill "..mission.kills, mission.color, "enemies")
			else
				self:AddObjective("box", "Kill " .. mission.kills .. " enemies")
			end
		end

		if (mission.survivors) then
			if (mission.color) then
				self:AddObjectiveWithColor("box", sprintf("Leave %.0f%% or fewer", mission.survivors*100), mission.color, "alive")
			else
				self:AddObjective("box", sprintf("Leave %.0f%% or fewer alive", mission.survivors*100))
			end
		end

		if (mission.bombs) then
			self:AddObjective("box", "Use no more than ".. mission.bombs .." bombs")
		end

		if (mission.exclude) then
			self:AddObjectiveWithColor("box", "Avoid", mission.exclude, "targets")
		end

		self:AddGap(1.5)

		self:AddMissionButtons()
		assert(self.accept and self.accept:IsShown())

	else -- no mission and not max-rank

		assert(false)

	end

	self:SetWidth(MARGIN + self:GetContentWidth() + MARGIN)
	-- print("MF width", self:GetWidth(), "height", self:GetHeight())
end


function MissionBriefing : OnChangeMission()
	-- choose a new random mission
	-- local mission = Addon:GetMission()

	local mission, more = Addon:GetMission()
	while (more and mission == self.mission) do
		mission, more = Addon:GetMission()
	end
	C_Timer.After(0.01, function() self:SetMission(mission) end)
end


--
--  Report
--

local MissionReport = Mixin({}, MissionFrame)


function Addon : NewMissionReport(mission, status)
	local report = Mixin(CreateFrame("FRAME"), MissionReport)
	report.shadows = true
	report:ReportMission(mission, status)
	return report
end


function MissionReport : ReportMission(mission, status)

	self:SetBackdrop({ edgeSize=18, edgeFile="Interface/DialogFrame/UI-DialogBox-Border", bgFile="Interface/QuestionFrame/question-background" })
	-- self:SetBackdropColor(0.8,0.8,0.8, 0.90)
	self:SetBackdropColor(1,1,1, 0.90)

	--[[
		After Action Report
		48 targets neutralized, 21 bombs used

		Mission Accomplished!
		_/ Kill 25 [RED] targets
		_/ Use no more than 15 bombs
		XX Avoid [GREEN] targets

		You've been promoted!
		Congratulations, Troublemaker!

		Strategic Objectives
		_/ Kill 2,500 enemy targets (2,531 killed)
		_/ Complete 25 missions (25 completed)
		[] Complete Troublemaker challenge mission
	--]]

	self:Clear()

	MARGIN = FONTSIZE * 1.5
	V_MARGIN = FONTSIZE * 1.5

	local title = self:AddLine("After Action Report", FONTSIZE+2)

	local underline = self:CreateTexture(nil, "ARTWORK")
	underline:SetColorTexture(0,0,0, 1)
	underline:SetSize(title.label1:GetUnboundedStringWidth()+4, 1)
	underline:SetPoint("TOPLEFT", title.label1, "BOTTOMLEFT", 0, 1)
	self:AddGap(3/FONTSIZE)

	-- self:AddLine(sprintf("Mission kills: %d   Bombs used: %d", status.totalkills, status.bombs))
	self:AddLine(sprintf("%d targets neutralized, %d bombs used", status.totalkills, status.bombs), FONTSIZE)

	local kills_before = Addon.DB.total_kills or 0
	local kills_after = kills_before + status.totalkills
	Addon.DB.total_kills = kills_after

	if (mission) then

		self:AddGap(1.0)
		local outcome = self:AddLine("Mission Accomplished!", FONTSIZE+1)	-- for size, for now
		self:AddGap(0.25)

		local success = true

		if (mission.kills) then
			if (mission.color) then
				local icon = (status.kills[mission.color] or 0) >= mission.kills and "succeeded" or "failed"
				self:AddObjectiveWithColor(icon, "Kill "..mission.kills, mission.color, "enemies")
				success = success and (icon == "succeeded")
			else
				local icon = status.totalkills >= mission.kills and "succeeded" or "failed"
				self:AddObjective(icon, "Kill " .. mission.kills .. " enemies")
				success = success and (icon == "succeeded")
			end
		end

		if (mission.survivors) then
			local killed = mission.color and (status.kills[mission.color] or 0) or status.totalkills
			local actual
			if (status.survivors == 0) then
				actual = 0
			elseif (killed == 0) then
				assert(status.survivors > 0)
				actual = 1
			else
				assert(status.survivors + killed > 0)
				actual = status.survivors / (killed + status.survivors)
			end
			local icon = actual <= mission.survivors and "succeeded" or "failed"
			if (mission.color) then
				self:AddObjectiveWithColor(icon, sprintf("Leave %.0f%% or fewer", mission.survivors*100), mission.color, "alive")
			else
				self:AddObjective(icon, sprintf("Leave %.0f%% or fewer alive", mission.survivors*100))
			end
			success = success and (icon == "succeeded")
		end

		if (mission.bombs) then
			local icon = (status.bombs <= mission.bombs) and "succeeded" or "failed"
			self:AddObjective(icon, "Use no more than ".. mission.bombs .." bombs")
			success = success and (icon == "succeeded")
		end

		if (mission.exclude) then
			local killed = status.kills[mission.exclude] or 0
			local icon = (killed == 0) and "succeeded" or "failed"
			self:AddObjectiveWithColor(icon, "Avoid", mission.exclude, "targets")
			success = success and (icon == "succeeded")
		end

		if (success) then
			outcome.label1:SetText(Addon.gold("Mission Accomplished!"))
			outcome.label1:SetShadowOffset(1,-1); outcome.label1:SetShadowColor(0,0,0, 1)
			C_Timer.After(0.33, function() PlaySound(37) end)	-- 37 is Resurrection, delay for game's bird-dismount sound
		else
			outcome.label1:SetText("Mission Failed.")
		end

		local missions_before = Addon.DB.total_missions or 0
		local missions_after = missions_before + (success and 1 or 0)
		Addon.DB.total_missions = missions_after

		local rank = Addon:GetCurrentRank()
		local promoted = false
		if (rank.promotion) then
			if (success and (mission == rank.promotion.challenge)) then
				local newrank = Addon:GetNextRank()
				-- print("promoting from", rank.name, "to", newrank.name)
				Addon.DB.rank = newrank.level
				assert(Addon:GetCurrentRank() == newrank) -- new rank now current
				rank = newrank
				promoted = true

				-- play a sound after a short delay to let the mission-accomplished should be heard
				local soundID
				if (UnitFactionGroup("player") == "Alliance") then
					-- 8455 PVPVictoryAlliance (Sky Marshall), or 8572 CrowdCheerAlliance2 (others)
					soundID = (newrank.level == MAX_RANK) and 8455 or 8572
				else
					-- 8454 PVPVictoryHorde (Sky Marshall), or 8573 CrowdCheerHorde1 (others)
					soundID = (newrank.level == MAX_RANK) and 8454 or 8573
				end
				C_Timer.After(0.33+0.75, function() PlaySound(soundID) end)

				self:AddGap(1.0)
				local line

				line = self:AddLine(Addon.gold("You've been promoted."))
				line.label1:SetShadowOffset(1,-1); line.label1:SetShadowColor(0,0,0, 1)

				line = self:AddLine(Addon.gold("Congratulations, " .. rank.name .. "!"))
				line.label1:SetShadowOffset(1,-1); line.label1:SetShadowColor(0,0,0, 1)

				self:AddGap(1.0)
				self:AddLine(sprintf("%s targets neutralized", Addon.commaformat(Addon.DB.total_kills)))
				self:AddLine(sprintf("%s missions completed", Addon.commaformat(Addon.DB.total_missions)))
				self:AddLine(sprintf("%s total sorties", Addon.commaformat(Addon.DB.total_sorties)))

				if (rank.level == Addon.MAX_RANK) then
					self:AddGap(1.0)
					line = self:AddLine(Addon.gold("Thank you for playing Azeroth Bombardier!"))
					line.label1:SetShadowOffset(1,-1); line.label1:SetShadowColor(0,0,0, 1)
				end
			end

			-- rank might have changed...
			if (rank.promotion) then
				self:AddGap(1.0)
				local prefix = promoted and "New " or ""
				self:AddLine(prefix.."Strategic Objectives", FONTSIZE+1)
				self:AddGap(0.25)

				local icon, text

				icon = (kills_after >= rank.promotion.kills) and "succeeded" or "box"
				-- text = sprintf("Kill %s enemies (%s killed)", Addon.commaformat(rank.promotion.kills), Addon.commaformat(kills_after))
				text = sprintf("Neutralized %s of %s targets", Addon.commaformat(kills_after), Addon.commaformat(rank.promotion.kills))
				self:AddObjective(icon, text)

				icon = (missions_after >= rank.promotion.missions) and "succeeded" or "box"
				-- text = sprintf("Complete %s missions (%s completed)", Addon.commaformat(rank.promotion.missions), Addon.commaformat(missions_after))
				text = sprintf("Completed %d of %d missions", Addon.commaformat(missions_after), Addon.commaformat(rank.promotion.missions))
				self:AddObjective(icon, text)

				-- also show the challenge mission?
				local nextrank = Addon:GetNextRank()
				if (mission == rank.promotion.mission) then
					if (success) then
						self:AddObjective("succeeded", sprintf("Complete %s challenge mission", nextrank.name))
					else
						self:AddObjective("box", sprintf("Complete %s challenge mission (Failed)", nextrank.name))
					end
				elseif (kills_after >= rank.promotion.kills and missions_after >= rank.promotion.missions) then
					self:AddObjective("box", sprintf("Complete %s challenge mission", nextrank.name))
				end
				-- else not enough kills/missions to reveal challenge mission yet
			end -- [new] rank has room for promotion

		end -- [original] rank has room for promotion

	else -- no mission

		self:AddGap(1.0)
		self:AddLine(sprintf("%s targets neutralized", Addon.commaformat(Addon.DB.total_kills)))
		self:AddLine(sprintf("%s missions completed", Addon.commaformat(Addon.DB.total_missions)))
		self:AddLine(sprintf("%s total sorties", Addon.commaformat(Addon.DB.total_sorties)))

	end -- no mission

	self:SetWidth(MARGIN + self:GetContentWidth() + MARGIN*1.5) -- why does this need more on the right when shown on the map??
	-- print("MF width", self:GetWidth(), "height", self:GetHeight())
end
